-----------------------------------------------------------------------------
--
-- data
--   \healingDone
--     \<healerName>
--       -count, amount, overhealAmount
--       \healDetails
--         \<healName>
--            -count, amount, overhealAmount
--			  \<healType> -- one of Heal, Critical Heal, Tick
--              -count, amount, overhealAmount
--   \healingTaken
--     \<healTargetName>
--       -count, amount, overhealAmount
--       \healDetails
--         \<healName>
--            -count, amount, overhealAmount
--			  \<healType> -- one of Heal, Critical Heal, Tick
--              -count, amount, overhealAmount
--
-----------------------------------------------------------------------------

local L = AceLibrary("AceLocale-2.2"):new("AssessmentHealing")
L:RegisterTranslations("enUS", function() return {
	["Healers List"] = true,
	["Healing Done"] = true,
	["HPS"] = true,
} end)

-----------------------------------------------------------------------------

local AssessmentHealing = Assessment:NewModule("Healing", "AceEvent-2.0", "Parser-3.0")
AssessmentHealing.db = Assessment:AcquireDBNamespace("Healing")
Assessment:RegisterDefaults("Healing", "profile", {
	enable = true
})

-----------------------------------------------------------------------------

Assessment.TagData.healingdone =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingDone = fightData.healingDone
				if not healingDone then return 0 end
				
				if not healingDone[data.key] then return 0 end
				return healingDone[data.key].amount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.overhealingdone =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingDone = fightData.healingDone
				if not healingDone then return 0 end
				
				if not healingDone[data.key] then return 0 end
				return healingDone[data.key].overhealAmount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.healingtaken =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingTaken = fightData.healingTaken
				if not healingTaken then return 0 end
				
				if not healingTaken[data.key] then return 0 end
				return healingTaken[data.key].amount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

Assessment.TagData.overhealingtaken =
{
	method = function(data)
				local fightData = data.fightData
				if not fightData then return 0 end
				
				local healingTaken = fightData.healingTaken
				if not healingTaken then return 0 end
				
				if not healingTaken[data.key] then return 0 end
				return healingTaken[data.key].overhealAmount
			 end,
	data = "healing",
	guaranteeResult = true,
	numeric = true,
}

-----------------------------------------------------------------------------

Assessment.ListType.Healers =
{
	method = function(fightData)
				if not fightData.healingDone then return function() return nil end end
				return pairs(fightData.healingDone)
			 end,
	data = "healing",
	description = L["Healers List"],
}

-----------------------------------------------------------------------------

Assessment.DefaultDisplaySets.HealingDone =
{
	titleTag		= "Healing Done",
	listType		= "GroupMembers", 
	barValueTag		= "healingDone", 
	filterTag		= "barValue ~= 0",
	fightIndex		= 0,
	text			= { { }, { tag= "[barValue:short]" } },
	description		= L["Healing Done"]
}

Assessment.DefaultDisplaySets.HPS =
{
	titleTag		= "HPS: [combattime:formatDuration]",
	listType		= "GroupMembers", 
	barValueTag		= "healingDone / maximum(combatTime, 1)", 
	filterTag		= "barValue ~= 0",
	fightIndex		= 1,
	text			= { { }, { tag= "[barValue:format(\"%.1f\")]" } },
	description		= L["HPS"]
}

-----------------------------------------------------------------------------

function AssessmentHealing:OnInitialize()
end

function AssessmentHealing:OnEnable()
	self:RegisterParserEvent({ eventType = "Heal" }, "OnParserHealEvent")
end

function AssessmentHealing:OnDisable()
end

-----------------------------------------------------------------------------

local function AccumulateData(data, eventData)
	data.count = data.count + 1
	data.amount = data.amount + eventData.amount
	data.overhealAmount = data.overhealAmount + eventData.overhealAmount
end

-----------------------------------------------------------------------------

function AssessmentHealing:NewDataTable()
	local data = self:NewTable()
	data.count = 0
	data.amount = 0
	data.overhealAmount = 0
	return data
end

-----------------------------------------------------------------------------

function AssessmentHealing:UpdateHealingData(dataTable, member, eventData, healType)
	local data = dataTable[member]
	if not data then
		data = self:NewDataTable()
		data.healDetails = self:NewTable()
		dataTable[member] = data
	end
	AccumulateData(data, eventData)
	
	local healDetailsParent = data.healDetails
	local healDetails = healDetailsParent[eventData.abilityName]
	if not healDetails then
		healDetails = self:NewDataTable()
		healDetailsParent[eventData.abilityName] = healDetails
	end
	AccumulateData(healDetails, eventData)
	
	local healTypeDetails = healDetails[healType]
	if not healTypeDetails then
		healTypeDetails = self:NewDataTable()
		healDetails[healType] = healTypeDetails
	end
	AccumulateData(healTypeDetails, eventData)
end

-----------------------------------------------------------------------------

function AssessmentHealing:OnParserHealEvent(eventData)
	local healType = "Heal"
	if eventData.isDoT then
		healType = "Tick"
	elseif eventData.isCrit then
		healType = "CriticalHeal"
	end
	
	for data in self:IterateUpdateDataTables() do
		data.dataUpdateTimes.healing = GetTime()
		local healingDone = data.healingDone
		if not healingDone then
			healingDone = self:NewTable()
			data.healingDone = healingDone
		end

		local healingTaken = data.healingTaken
		if not healingTaken then
			healingTaken = self:NewTable()
			data.healingTaken = healingTaken
		end

		self:UpdateHealingData(healingDone, eventData.sourceName, eventData, healType)
		self:UpdateHealingData(healingTaken, eventData.recipientName, eventData, healType)
	end
end

-----------------------------------------------------------------------------
