-----------------------------------------------------------------------------
--
-- Per window configuration:
--  Title
--  Title appearance
--  Border
--  Background color
--  BarValue
--  SortCriteria
--  Text Tags
--  Bar appearance
--
-----------------------------------------------------------------------------

local AceOO = AceLibrary("AceOO-2.0")
local SML = AceLibrary("SharedMedia-1.0")
local Dewdrop = AceLibrary("Dewdrop-2.0")
local AssessmentUserInterface = Assessment:NewModule("UserInterface", "AceEvent-2.0")

-----------------------------------------------------------------------------

AssessmentUserInterface.db = Assessment:AcquireDBNamespace("UserInterface")

local profileDefaults =
{
	windows =
	{
-- Set on first login if it doesn't exist, and defaultWindowProfile copied in
--		count = 1,
	},
	displaySets =
	{
		-- Defaults are inserted here in OnInitialize
	}
}

local defaultWindowProfile =
{
	displaySetStack		= { "Assessment" },
	displaySetStackLevel= 1,
	updateRate			= 10,
	title = 
	{
		enable			= true,
		alignment		= "CENTER",
		font			= "Arial Narrow",
		fontSize		= 9,
		outline			= "OUTLINE",
		shadow			= true,
		height			= 10,
		textColor		= { 1, 1, 1, 1 },
		backgroundColor	= { 0, 0, 0, 1 }
	},
	bar = 
	{ 
		count			= 10,
		texture			= "Blizzard",
		defaultColor	= { 0.75, 0.75, 0.75 },
		highlightColor	= { 1, 1, 0, 0.35 },
	},
	frame = 
	{ 
		lock			= false,
		width			= 110,
		height			= 140, 
		point			= "CENTER",
		relativePoint	= "CENTER",
		x				= 0, 
		y				= 0, 
		borderStyle		= "Default",
		borderColor		= { 1, 1, 1, 0.75 },
		backdropColor	= { 0, 0, 0, 0.75 },
		visible			= true,
	},
}

local defaultDisplaySetProfile =
{
	titleTag		= "Assessment",
	
	listType		= "GroupMembers", 
	barValueTag		= "damageDealt", 
	sortKeyTag		= "barValue",
	filterTag		= "barValue ~= 0",
	descendingSort	= true,
	fightIndex		= 0,
	colorTypeList	= { "Class", count = 1 },
	favorite		= true,
	selectable		= true,
	
	onClickType		= "NONE",		-- NONE, or PUSH_DISPLAY_SET currently.
--	onClickType		= "PUSH_DISPLAY_SET",
--  onClickDisplaySet = "Display Set Name",
--	onClickData		= "EnumerateTable|damageDealt|[key]|attackDetails",
	
	text =
	{
		count = 2,
		{
			alignment	= "LEFT",
			font		= "Arial Narrow",
			fontSize	= 9,
			outline		= "",
			shadow		= true,
			alpha		= 1,
			tag			= "[rank]. [key]",
			offsets =
			{
				left		= 0,
				right		= -20,
				vertical	= 0,
			},
		},
		{
			alignment	= "RIGHT",
			font		= "Arial Narrow",
			fontSize	= 9,
			outline		= "",
			shadow		= true,
			alpha		= 1,
			tag			= "[barValue]",
			offsets =
			{
				left		= 0,
				right		= 0,
				vertical	= 0,
			},
		},
	},
}

local defaultTextProfile =
{
	alignment	= "LEFT",
	font		= "Arial Narrow",
	fontSize	= 9,
	outline		= "",
	shadow		= true,
	alpha		= 1,
	tag			= "text",
	offsets =
	{
		left		= 0,
		right		= 0,
		vertical	= 0,
	},
}

-----------------------------------------------------------------------------

-- This thing is created on demand, and reparented as necessary
local uiBarHighlightFrame

local UIBar = AceOO.Class()

local function OnUIBarEnter(self, motion)
	local frame = self.uiBar.frame
	if not uiBarHighlightFrame then
		uiBarHighlightFrame = frame:CreateTexture(nil, "HIGHLIGHT")
		uiBarHighlightFrame:SetTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
		uiBarHighlightFrame:SetBlendMode("ADD")
	end
	uiBarHighlightFrame:SetVertexColor(unpack(self.uiBar.windowOptions.bar.highlightColor))
	uiBarHighlightFrame:SetParent(frame)
	uiBarHighlightFrame:ClearAllPoints()
	uiBarHighlightFrame:SetAllPoints(frame)
	uiBarHighlightFrame:Show()
end

local function OnUIBarLeave(self, motion)
	uiBarHighlightFrame:Hide()
end

local function OnUIBarMouseDown(self, button)
	if button == "LeftButton" then
		self:GetParent().window:OnBarClicked(self.uiBar)
	elseif button == "RightButton" then
		self:GetParent().window:PopDisplaySetStack()
	end
end

function UIBar.prototype:init(parent, windowOptions, displaySet, index)
	UIBar.super.prototype.init(self)

	local frame = CreateFrame("Frame", nil, parent)
	frame.uiBar = self
	self.frame = frame

	local bar = frame:CreateTexture()
	self.bar = bar
	self.texts = { }
	self.barColorMethods = { }

	bar:SetPoint("TOPLEFT", frame, "TOPLEFT")
	bar:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT")
	
	frame:SetScript("OnEnter", OnUIBarEnter)
	frame:SetScript("OnLeave", OnUIBarLeave)
	frame:SetScript("OnMouseDown", OnUIBarMouseDown)
	frame:EnableMouse()

	self:SetOptions(parent, windowOptions, displaySet, index)
end

function UIBar.prototype:SetOptions(parent, windowOptions, displaySet, index)
	self.windowOptions = windowOptions
	self.displaySet = displaySet

	local bar = self.bar

	for i in ipairs(self.barColorMethods) do
		self.barColorMethods[i] = nil
	end
	for i = 1, displaySet.colorTypeList.count do
		self.barColorMethods[i] = Assessment.ColorType[displaySet.colorTypeList[i]].method
	end
	self.barColorMethods[#self.barColorMethods+1] = function() return windowOptions.bar.defaultColor end

	bar:SetTexture(SML:Fetch("statusbar", windowOptions.bar.texture))
	bar:SetVertexColor(1, 1, 1, 1)

	for i = 1, self.displaySet.text.count do
		if not self.texts[i] then
			self.texts[i] = self.frame:CreateFontString()
		end
		self.texts[i]:Show()
	end

	for i = self.displaySet.text.count + 1, #self.texts do
		self.texts[i]:Hide()
	end
	
	self:PlaceBar(parent, index)
end

function UIBar.prototype:PlaceBar(parent, index)
	local frame = self.frame
	local bar = self.bar
	local verticalOffset = -5 - (self.windowOptions.title.enable and self.windowOptions.title.height or 0)
	local height = (parent:GetHeight()+verticalOffset-5) / self.windowOptions.bar.count
	
	self.maximumBarWidth = (self.windowOptions.frame.width-10)
	
	frame:ClearAllPoints()
	frame:SetPoint("TOPLEFT", parent, "TOPLEFT", 5, -(index-1)*height+verticalOffset)
	frame:SetPoint("BOTTOMRIGHT", parent, "TOPRIGHT", -5, -index*height+verticalOffset)
	
	for i = 1, self.displaySet.text.count do
		local text = self.texts[i]
		local options = self.displaySet.text[i]
		
		text:ClearAllPoints()
		text:SetPoint("LEFT", frame, "LEFT", options.offsets.left)
		text:SetPoint("RIGHT", frame, "RIGHT", options.offsets.right)
		text:SetPoint("TOP", frame, "TOP", 0, options.offsets.vertical)
		text:SetPoint("BOTTOM", frame, "BOTTOM", 0, options.offsets.vertical)

		text:SetAlpha(options.alpha)

		text:SetFont(SML:Fetch("font", options.font), options.fontSize, options.outline)
		if options.shadow then
			text:SetShadowOffset(1, -1)
		else
			text:SetShadowOffset(0, 0)
		end
		text:SetJustifyH(options.alignment)
	end
end

function UIBar.prototype:SetFraction(fraction)
	local width = fraction * self.maximumBarWidth
	local bar = self.bar
	if width == 0 then
		bar:Hide()
	else
		bar:SetWidth(width)
		bar:Show()
	end
end

function UIBar.prototype:Hide()
	self.frame:Hide()
end

function UIBar.prototype:Update(data, barNormalizeMultiplier, textMethods)
	-- Update Bar
	self:SetFraction(data.barValue * barNormalizeMultiplier)
	
	local color
	for i,f in ipairs(self.barColorMethods) do
		color = f(data.fightData, data.key)
		if color then break end
	end
	self.bar:SetVertexColor(unpack(color))
	
	-- Update Texts
	for i = 1, self.displaySet.text.count do
		local text = self.texts[i]
		text:SetText(textMethods[i](data))
	end
	self.frame:Show()
end

-----------------------------------------------------------------------------

local borderStyles =
{
	None		= { edgeFile = "", inset = 5 },
	Default		= { edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", inset = 5 },
	Textured	= { edgeFile = "", inset = 3 },
	Square		= { edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border", inset = 5 },
}

local UIWindow = AceOO.Class("AceEvent-2.0")

function UIWindow.prototype:init(windowOptions, displaySet)
	UIWindow.super.prototype.init(self)
	local frame = CreateFrame("Frame", nil, UIParent)
	frame.window = self

	self.dependentData		= { }
	self.dataUpdateTimes	= { }
	self.textMethods		= { }
	self.listTypeParameters = { }

	self.frame			= frame
	self.bars			= { }
	
	self.titleText = frame:CreateFontString()
	self.titleTexture = frame:CreateTexture()

	self:SetWindowOptions(windowOptions)
end

local bgTable = 
{
    bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
    tile = true,
    tileSize = 16,
    edgeSize = 16,
	insets = { left = inset, right = inset, top = inset, bottom = inset }
}

local isWindowSizing, wasWindowDragged, wasWindowSizeChanged

local function OnUIWindowMouseDown(this, button)
	if button == "LeftButton" then
		if not this.window.windowOptions.frame.lock then 
			local mouseDownX, mouseDownY = GetCursorPosition(UIParent);

			local effectiveScale = this:GetEffectiveScale();
			local left = this:GetLeft() * effectiveScale;
			local top = this:GetTop() * effectiveScale;
			local right = this:GetRight() * effectiveScale;
			local bottom = this:GetBottom() * effectiveScale;
			
			local edge = "";
			if (abs(mouseDownY - bottom) < 7) then
				edge = "BOTTOM";
			elseif (abs(mouseDownY - top) < 7) then
				edge = "TOP"
			end
			
			if (abs(mouseDownX - left) < 7) then
				edge = edge.."LEFT";
			elseif (abs(mouseDownX - right) < 7) then
				edge = edge.."RIGHT";
			end
			
			if edge ~= "" then
				this:StartSizing(edge)
				isWindowSizing = true
				wasWindowSizeChanged = nil
			end
		end
	elseif button == "RightButton" then
		this.window:PopDisplaySetStack()
	end
end

local function OnUIWindowMouseUp(this, button)
	if button ~= "LeftButton" then return end
	
	if isWindowSizing then
		this:StopMovingOrSizing() 
		isWindowSizing = nil
		if wasWindowSizeChanged then
			wasWindowSizeChanged = nil
		else
			this.window:OnClick(button)
		end
	elseif wasWindowDragged then
		wasWindowDragged = nil
	else
		this.window:OnClick(button)
	end
end

local function OnUIWindowDragStart(this)
	wasWindowDragged = true
	if not this.window.windowOptions.frame.lock then 
		this:StartMoving() 
	end
end

local function OnUIWindowDragStop(this)
	this:StopMovingOrSizing() 
	local point,_,relativePoint,x,y = this:GetPoint(1)
	local frameOptions = this.window.windowOptions.frame
	frameOptions.point = point
	frameOptions.relativePoint = relativePoint
	frameOptions.x = x
	frameOptions.y = y
end

local function OnUIWindowSizeChanged(this, width, height)
	wasWindowSizeChanged = true
	local window = this.window
	local frameOptions = window.windowOptions.frame
	frameOptions.width = width 
	frameOptions.height = height

	local point,_,relativePoint,x,y = this:GetPoint(1)
	frameOptions.point = point
	frameOptions.relativePoint = relativePoint
	frameOptions.x = x
	frameOptions.y = y

	window:LayoutBars()
	window.forceUpdate = true
	window:Update()
end

function UIWindow.prototype:SetWindowOptions(windowOptions)
	local displaySet = AssessmentUserInterface:GetDisplaySet(windowOptions)
	local _, displaySetParameterText = AssessmentUserInterface:GetDisplaySetName(windowOptions)

	-- Clear some tables
	for k in pairs(self.dependentData) do self.dependentData[k] = nil end
	for k in pairs(self.dataUpdateTimes) do self.dataUpdateTimes[k] = nil end
	for k in ipairs(self.textMethods) do self.textMethods[k] = nil end

	self.topBarOffset	= 0
	self.forceUpdate	= true
	self.displaySet		= displaySet
	self.windowOptions	= windowOptions

	self.listTypeMethod = Assessment.ListType[displaySet.listType].method
	Assessment:ReleaseTable(self.listTypeParameters)
	self.listTypeParameters = displaySetParameterText and { (","):split(displaySetParameterText) } or Assessment:NewTable()
	
	local compiledFilterValue = Assessment:CompileTag(displaySet.filterTag, "raw")
	self:MergeRequiredData(compiledFilterValue.data)
	self.filterMethod = compiledFilterValue.method

	for i=1,displaySet.colorTypeList.count do
		self:MergeRequiredData(Assessment.ColorType[displaySet.colorTypeList[i]].data)
	end
	self:MergeRequiredData(Assessment.ListType[displaySet.listType].data)

	if displaySet.descendingSort then
		self.sortDirectionMethod = function(a, b) if a.sortValue == b.sortValue then
													return a.key < b.key	-- always sort keys ascending if the sort values are equal
												  else
												    return a.sortValue > b.sortValue
												  end
								   end
	else
		self.sortDirectionMethod = function(a, b) if a.sortValue == b.sortValue then
													return a.key < b.key
												  else
												    return a.sortValue < b.sortValue
												  end
								   end
	end

	local compiledBarValue = Assessment:CompileTag(displaySet.barValueTag, "number")
	self.barValueMethod = compiledBarValue.method
	self:MergeRequiredData(compiledBarValue.data)
	
	for i=1,displaySet.text.count do
		local compiledTextValue = Assessment:CompileTag(displaySet.text[i].tag, "string")
		self.textMethods[i] = compiledTextValue.method
		self:MergeRequiredData(compiledTextValue.data)
	end

	local compiledSortValue = Assessment:CompileTag(displaySet.sortKeyTag, "raw")
	self:MergeRequiredData(compiledSortValue.data)
	self.sortMethod = compiledSortValue.method

	-------

	local frame = self.frame
	local frameOptions = windowOptions.frame
	frame:ClearAllPoints()
	frame:SetPoint(frameOptions.point, UIParent, frameOptions.relativePoint, frameOptions.x, frameOptions.y)
	frame:SetWidth(frameOptions.width)
	frame:SetHeight(frameOptions.height)
	frame:SetMovable(true)
	frame:SetResizable(true)
	frame:SetMinResize(32,32)
	frame:EnableMouse(true)
	frame:EnableMouseWheel(true)
	frame:RegisterForDrag("LeftButton")

	frame:SetScript("OnMouseWheel", function(this, delta) 
										local window = this.window
										if delta < 0 then
											window.topBarOffset = window.topBarOffset + 1
										else
											window.topBarOffset = window.topBarOffset - 1
											if window.topBarOffset < 0 then
												window.topBarOffset = 0
											end
										end
										window.forceUpdate = true
										window:Update()
									end)
	
	frame:SetScript("OnMouseDown",	OnUIWindowMouseDown)
	frame:SetScript("OnMouseUp",	OnUIWindowMouseUp)
	frame:SetScript("OnDragStart",	OnUIWindowDragStart)
	frame:SetScript("OnDragStop",	OnUIWindowDragStop)
	frame:SetScript("OnSizeChanged",OnUIWindowSizeChanged)

	local borderStyle = borderStyles[windowOptions.frame.borderStyle]
	bgTable.edgeFile = borderStyle.edgeFile
	local inset = borderStyle.inset
	local insets = bgTable.insets
	insets.left = inset
	insets.top = inset
	insets.right = inset
	insets.bottom = inset
    frame:SetBackdrop(bgTable)
	frame:SetBackdropBorderColor(unpack(windowOptions.frame.borderColor))
	frame:SetBackdropColor(unpack(windowOptions.frame.backdropColor))

	--------

	local titleText = self.titleText
	local titleOptions = windowOptions.title
	
	if titleOptions.enable then
		titleText:ClearAllPoints()
		titleText:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		titleText:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
		titleText:SetTextColor(unpack(titleOptions.textColor))

		titleText:SetFont(SML:Fetch("font", titleOptions.font), titleOptions.fontSize, titleOptions.outline)
		if titleOptions.shadow then
			titleText:SetShadowOffset(1, -1)
		else
			titleText:SetShadowOffset(0, 0)
		end
		titleText:SetJustifyH(titleOptions.alignment)

		local compiledTitleTextValue = Assessment:CompileTag(displaySet.titleTag, "string")
		self.titleTextMethod = compiledTitleTextValue.method
		self:MergeRequiredData(compiledTitleTextValue.data)

		titleText:Show()

		local titleTexture = self.titleTexture
		titleTexture:ClearAllPoints()
		titleTexture:SetPoint("TOPLEFT", frame, "TOPLEFT", 5, -5)
		titleTexture:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -5, -titleOptions.height-5)
		titleTexture:SetTexture(unpack(titleOptions.backgroundColor))
		titleTexture:Show()
	else
		titleText:Hide()
		self.titleTexture:Hide()
	end

	--------

	for i = 1, windowOptions.bar.count do
		if self.bars[i] then
			self.bars[i]:SetOptions(frame, windowOptions, displaySet, i)
		else
			self.bars[i] = UIBar:new(frame, windowOptions, displaySet, i)
		end
	end
	for i = windowOptions.bar.count+1, #self.bars do
		self.bars[i]:Hide()
	end

	if frameOptions.visible then
		self:Show()
	else
		self:Hide()
	end
end

function UIWindow.prototype:GetBarIndex(bar)
	for i = 1, self.windowOptions.bar.count do
		if self.bars[i] == bar then
			return i
		end
	end
	error("Internal error: Bar not found")
end

function UIWindow.prototype:PopDisplaySetStack(windowOptions)
	AssessmentUserInterface:PopWindowDisplaySet(self)
end

function UIWindow.prototype:OnClick(buttonName)
	Dewdrop:Open("",
					'children', function() self:UpdateDewdrop() end,
					'cursorX', true,
					'cursorY', true
					)
end

function UIWindow.prototype:UpdateDewdrop()
	local currentDisplaySetName = AssessmentUserInterface:GetDisplaySetName(self.windowOptions)
	local sortList = Assessment:NewTable()
	
	for k,v in pairs(AssessmentUserInterface.db.profile.displaySets) do
		if k == currentDisplaySetName or (v.favorite and v.selectable) then
			table.insert(sortList, k)
		end
	end
	table.sort(sortList)

	for i,k in ipairs(sortList) do
		Dewdrop:AddLine('text', k,
					'checked', k == currentDisplaySetName,
					'func', k ~= currentDisplaySetName and "SetDisplaySet" or nil,
					'arg1', AssessmentUserInterface,
					'arg2', self,
					'arg3', k)
	end

	Assessment:ReleaseTable(sortList)
end

function UIWindow.prototype:Show()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
	end

	local windowOptions = self.windowOptions
	local updateRate = windowOptions.updateRate ~= 0 and 1/windowOptions.updateRate or 0

	self.updater = self:ScheduleRepeatingEvent(self.Update, updateRate, self)
	self.frame:Show()
	self:Update()
end

function UIWindow.prototype:Hide()
	if self.updater then
		self:CancelScheduledEvent(self.updater)
		self.updater = nil
	end
	self.frame:Hide()
end

function UIWindow.prototype:LayoutBars(fightData)
	for i=1, self.windowOptions.bar.count do
		self.bars[i]:PlaceBar(self.frame, i)
	end
end

function UIWindow.prototype:MergeRequiredData(data)
	if not data then return end
	if type(data) == "string" then
		self.dependentData[data] = true
	elseif type(data) == "table" then
		for k in pairs(data) do
			self.dependentData[k] = true
		end
	end
end

function UIWindow.prototype:NeedsUpdate(fightData)
	if self.dependentData.ALWAYSUPDATE then
		return true
	end

	local dataUpdateTimes = self.dataUpdateTimes
	if not fightData then
		if next(dataUpdateTimes) then
			for k in pairs(dataUpdateTimes) do
				dataUpdateTimes[k] = nil
			end
			return true
		end
		if self.forceUpdate then
			self.forceUpdate = nil
			return true
		end
		return false
	end

	local result = false
	for k in pairs(self.dependentData) do
		if dataUpdateTimes[k] ~= fightData.dataUpdateTimes[k] then
			result = true
			dataUpdateTimes[k] = fightData.dataUpdateTimes[k]
		end
	end
	if self.forceUpdate then
		self.forceUpdate = nil
		return true
	end
	return result
end

local sortList = { }
local maximumBarValue
local totalBarValue

function UIWindow.prototype:UpdateSortList(fightData)
	maximumBarValue = 0
	totalBarValue = 0
	for i,v in ipairs(sortList) do
		sortList[i] = Assessment:ReleaseTable(v)
	end

	if fightData then
		for k,v in self.listTypeMethod(fightData, unpack(self.listTypeParameters)) do
			local data = Assessment:NewTable()
			data.key = k
			data.value = v
			data.fightData = fightData
			
			local barValue = self.barValueMethod(data)
			data.barValue = barValue
			
			if self.filterMethod(data) then
				if barValue > maximumBarValue then
					maximumBarValue = barValue
				end
				totalBarValue = totalBarValue + barValue

				data.barValue = barValue
				sortList[#sortList+1] = data
			else
				Assessment:ReleaseTable(data)
			end			
		end
		
		for _,data in ipairs(sortList) do
			data.totalBarValue = totalBarValue
			data.maximumBarValue = maximumBarValue
			data.sortValue = self.sortMethod(data)
		end

		table.sort(sortList, self.sortDirectionMethod)
	end
end

Assessment.TagData.windowdisplaysetparameter =
{
	method = function(data, value, index)
				return value and data.window.listTypeParameters[index]
			 end,
	parameterTypes = { "number" }
}

local titleData = { key = UnitName("player"), rank = 0 }
function UIWindow.prototype:DoUpdate(fightData)
	self:UpdateSortList(fightData)

	local barNormalizeMultiplier = maximumBarValue ~= 0 and 1 / maximumBarValue or 0

	local topBarOffset = self.topBarOffset
	local barCount = self.windowOptions.bar.count
	if topBarOffset + barCount > #sortList then
		topBarOffset = #sortList - barCount
		if topBarOffset < 0 then
			topBarOffset = 0
		end
		self.topBarOffset = topBarOffset
	end

	local endDisplay = barCount
	if endDisplay > #sortList-topBarOffset then 
		endDisplay = #sortList-topBarOffset 
		for i = endDisplay+1, barCount do
			self.bars[i]:Hide()
		end
	end
	for i = 1, endDisplay do
		local rank = i+topBarOffset
		local data = sortList[rank]
		data.rank = rank
		self.bars[i]:Update(data, barNormalizeMultiplier, self.textMethods)
	end	

	if self.windowOptions.title.enable then
		titleData.fightData = fightData
		titleData.totalBarValue = totalBarValue
		titleData.maximumBarValue = maximumBarValue
		titleData.window = self

		self.titleText:SetText( self.titleTextMethod(titleData) )
	end
end

function UIWindow.prototype:Update()
	local fightData = Assessment:GetFightData(self.displaySet.fightIndex)
	if not self:NeedsUpdate(fightData) then return end
	self:DoUpdate(fightData)
end

function UIWindow.prototype:OnBarClicked(bar)
	if self.displaySet.onClickType == "PUSH_DISPLAY_SET" and self.displaySet.onClickDisplaySet and self.displaySet.onClickData then
		local barIndex = self:GetBarIndex(bar)
		
		self:UpdateSortList(Assessment:GetFightData(self.displaySet.fightIndex))
		local data = sortList[barIndex+self.topBarOffset]
		
		if data then
			local displaySet = Assessment:CompileTag(self.displaySet.onClickDisplaySet, "string").method(data)
			local parameters = Assessment:CompileTag(self.displaySet.onClickData, "string").method(data)
			AssessmentUserInterface:PushWindowDisplaySet(self, displaySet.."$"..parameters)
		end
	end
end

-----------------------------------------------------------------------------

local function CopyTable(source)
	local result = Assessment:NewTable()

	for k,v in pairs(source) do
		if type(v) == "table" then
			result[k] = CopyTable(v)
		else
			result[k] = v
		end
	end
	
	return result
end

local function MergeTable(destination, source)
	for k,v in pairs(source) do
		if type(v) == "table" then
			MergeTable(destination[k], v)
		else
			destination[k] = v
		end
	end
end

function AssessmentUserInterface:OnInitialize()
	self.windows = { }

	for k,v in pairs(Assessment.DefaultDisplaySets) do
		local displaySet = CopyTable(defaultDisplaySetProfile)
		MergeTable(displaySet, v)
		
		-- Cleanup fields
		displaySet.description = nil
		
		profileDefaults.displaySets[v.description] = displaySet
	end
	
	Assessment:RegisterDefaults("UserInterface", "profile", profileDefaults)
	
	if not self.db.profile.windows.count then
		self.db.profile.windows.count = 1
		self.db.profile.windows[1] = CopyTable(defaultWindowProfile)
	end
	
	self:CreateValidateLists()
end

function AssessmentUserInterface:OnEnable()
	self:RegisterEvent('SharedMedia_SetGlobal', function(ty, val)
		if ty == "statusbar" then
			for k,v in ipairs(self.windows) do
				for k,v in ipairs(v.bars) do
					v.bar:SetTexture(SML:Fetch(ty, val))
				end
			end
		end
	end)
	self:CreateWindows()
	self:CreateOptions()
end

function AssessmentUserInterface:OnDisable()
end

function AssessmentUserInterface:CreateWindows()
	for i = 1, self.db.profile.windows.count do
		local windowOptions = self.db.profile.windows[i]
		if self.windows[i] then
			self.windows[i]:SetWindowOptions(windowOptions)
		else
			self.windows[i] = UIWindow:new(windowOptions)
		end
	end
	
	for i = self.db.profile.windows.count+1, #self.windows do
		self.windows[i]:Hide()
	end
end

function AssessmentUserInterface:GetWindowIndex(window)
	for i = 1,self.db.profile.windows.count do
		if self.windows[i] == window then
			return i
		end
	end
	error("INTERNAL ERROR: Window not found")
end

function AssessmentUserInterface:SetDisplaySet(window, displaySetName)
	Dewdrop:Close()
	local windowIndex = self:GetWindowIndex(window)
	local windowOptions = self.db.profile.windows[windowIndex]
	windowOptions.displaySetStackLevel = 1
	windowOptions.displaySetStack[1] = displaySetName
	self:UpdateWindowOptions(windowIndex)
end

function AssessmentUserInterface:PushWindowDisplaySet(window, displaySetName)
	local windowIndex = self:GetWindowIndex(window)
	local windowOptions = self.db.profile.windows[windowIndex]
	
	if not self.db.profile.displaySets[("$"):split(displaySetName, 2)] then
		error("Unable to set display set: "..displaySetName)
	end
	
	windowOptions.displaySetStackLevel = windowOptions.displaySetStackLevel + 1
	windowOptions.displaySetStack[windowOptions.displaySetStackLevel] = displaySetName
	self:UpdateWindowOptions(windowIndex)
end

function AssessmentUserInterface:PopWindowDisplaySet(window)
	local windowIndex = self:GetWindowIndex(window)
	local windowOptions = self.db.profile.windows[windowIndex]
	
	if windowOptions.displaySetStackLevel == 1 then return end
	
	windowOptions.displaySetStack[windowOptions.displaySetStackLevel] = nil
	windowOptions.displaySetStackLevel = windowOptions.displaySetStackLevel - 1
	self:UpdateWindowOptions(windowIndex)
end

function AssessmentUserInterface:GetDisplaySetName(windowOptions)
	return ("$"):split(windowOptions.displaySetStack[windowOptions.displaySetStackLevel], 2)
end

function AssessmentUserInterface:GetDisplaySet(windowOptions)
	return self.db.profile.displaySets[self:GetDisplaySetName(windowOptions)]
end

function AssessmentUserInterface:AddDisplaySet(displaySetName)
	if displaySetName then
		self.db.profile.displaySets[displaySetName] = CopyTable(defaultDisplaySetProfile)
		self:CreateOptions()
	else
		if Dewdrop then Dewdrop:Close() end
		StaticPopup_Show("ASSESSMENT_ADD_DISPLAY_SET")
	end
end

function AssessmentUserInterface:AddWindow()
	local windowOptions = self.db.profile.windows
	windowOptions.count = windowOptions.count + 1
	
	if not windowOptions[windowOptions.count] then
		windowOptions[windowOptions.count] = CopyTable(defaultWindowProfile)
	end
	
	self:CreateWindows()
	self:CreateOptions()
end

function AssessmentUserInterface:AddText(displaySetName)
	local text = self.db.profile.displaySets[displaySetName].text
	text.count = text.count + 1
	if not text[text.count] then
		text[text.count] = CopyTable(defaultTextProfile)
	end

	self:UpdateDisplaySets(displaySetName)
	self:CreateOptions()
end

-----------------------------------------------------------------------------
